import datetime
import os
import re
import sysconfig
import time
from importlib.metadata import (
    Distribution,
    distributions,
)
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Set,
    Tuple,
)


SITE_PACKAGES_PATH = sysconfig.get_paths()["platlib"]


def generate_pip_tree(path: str = SITE_PACKAGES_PATH) -> Tuple[List[Dict[str, Any]], int]:
    """Generate the Pip Tree of the virtual environment specified."""
    pip_tree_results = []
    required_by_data: Dict[str, Set[str]] = {}
    package_count = 0

    packages = get_pip_package_list(path)

    for package_count, package in enumerate(packages, start=1):
        package_details = get_package_details(package)
        _generate_reverse_requires_field(required_by_data, package_details)
        pip_tree_results.append(package_details)

    # Append the `required_by` field to each record created from `_generate_reverse_requires_field()`
    for item in pip_tree_results:
        item["required_by"] = sorted(required_by_data.get(item["name"], []))

    final_output = sorted(pip_tree_results, key=lambda k: k["name"].lower())

    return final_output, package_count


def get_pip_package_list(path: str = SITE_PACKAGES_PATH) -> Iterable[Distribution]:
    """Get the Pip package list of a Python virtual environment.

    Must be a path like: /project/venv/lib/python3.12/site-packages
    """
    packages = distributions(path=[path])

    return packages


def get_package_details(package: Distribution) -> Dict[str, Any]:
    """Build a dictionary of details for a package from Pip."""
    package_location = package._path  # type:ignore
    package_updated_at = (
        time.ctime(os.path.getctime(package_location))
        if package_location and os.path.exists(package_location)
        else "unknown"
    )

    requires_list = (
        [sorted(str(requirement.replace(" ", "").split(";")[0]) for requirement in package.requires)]
        if package.requires
        else []
    )

    package_details = {
        "name": package.metadata["Name"],
        "version": package.metadata["Version"],
        "updated": (
            datetime.datetime.strptime(package_updated_at, "%a %b %d %H:%M:%S %Y").strftime("%Y-%m-%d")
            if package_updated_at != "unknown"
            else "unknown"
        ),
        "requires": sorted([item for sublist in requires_list for item in set(sublist)]),
    }

    return package_details


def _generate_reverse_requires_field(required_by_data: Dict[str, Set[str]], package_details: Dict[str, Any]):
    """Generate a reversed list from the `requires` fields and create a
    collection of each `required_by` fields so each package can show what it's required by.
    """
    requires_list = [item for item in package_details["requires"]]
    for required_by_package in requires_list:
        word = re.compile(r"^(\w)+")
        name_match = word.match(required_by_package)
        required_by_package_name = name_match.group() if name_match is not None else ""

        # If a package is listed, add to its set, otherwise create a new set
        if required_by_package_name in required_by_data:
            required_by_data[required_by_package_name].add(package_details["name"])
        else:
            required_by_data[required_by_package_name] = {package_details["name"]}

    required_by_data = {key: set(value) for key, value in required_by_data.items()}



# flake8: noqa
from pip_tree.tree import (
    SITE_PACKAGES_PATH,
    generate_pip_tree,
    get_package_details,
    get_pip_package_list,
)


__all__ = [
    "generate_pip_tree",
    "get_package_details",
    "get_pip_package_list",
]



import argparse
import json
import os

import pip_tree
from pip_tree._version import __version__


class PipTreeCli:
    def __init__(self):
        parser = argparse.ArgumentParser(
            description="Get the dependency tree of your Python virtual environment via Pip."
        )
        parser.add_argument(
            "-p",
            "--path",
            required=False,
            default=pip_tree.SITE_PACKAGES_PATH,
            type=str,
            help=(
                "The path to the site-packages directory of a Python virtual environment. If a path is not provided,"
                " the virtual environment Pip Tree is run from will be used."
            ),
        )
        parser.add_argument(
            "--version",
            action="version",
            version=f"%(prog)s {__version__}",
        )
        parser.parse_args(namespace=self)

    def generate_console_output(self):
        """Take the output of the dependency tree and print to console."""
        print("Generating Pip Tree report...")

        final_output, package_count = pip_tree.generate_pip_tree(os.path.expanduser(self.path))
        console_output = json.dumps(final_output, indent=4)

        print(console_output)
        print(f'Pip Tree report complete! {package_count} dependencies found for "{self.path}".')


def main():
    PipTreeCli().generate_console_output()


if __name__ == "__main__":
    main()



import pip_tree


def test_get_package_details():
    """The poor man's approach to unit testing this library is to feed
    its own virtual environment details to itself and assert they match.

    Asserting the entire collection is difficult due to odd environment
    dependencies (such as pip itself being a different version in build systems),
    so we simply assert that certain attributes exist for a package such as `pytest`.
    """
    package_details, package_count = pip_tree.generate_pip_tree()

    assert any(item["name"] == "pytest" for item in package_details)
    assert any([] == item["requires"] for item in package_details)
    assert any("pytest" in item["required_by"] for item in package_details)
    assert 25 < package_count < 50  # Assert the package count is within a reasonable number for this package



import re

import setuptools

with open('README.md', 'r') as readme_file:
    long_description = readme_file.read()

# Inspiration: https://stackoverflow.com/a/7071358/6064135
with open('pip_tree/_version.py', 'r') as version_file:
    version_groups = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file.read(), re.M)
    if version_groups:
        version = version_groups.group(1)
    else:
        raise RuntimeError('Unable to find version string!')

DEV_REQUIREMENTS = [
    'bandit == 1.9.*',
    'black == 25.*',
    'build == 1.3.*',
    'flake8 == 7.*',
    'isort == 7.*',
    'mypy == 1.18.*',
    'pytest == 9.*',
    'pytest-cov == 7.*',
    'setuptools >= 80',
]

setuptools.setup(
    name='pip-tree',
    version=version,
    description='Get the dependency tree of your Python virtual environment via Pip.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/justintime50/pip-tree',
    author='Justintime50',
    license='MIT',
    packages=setuptools.find_packages(),
    package_data={'pip_tree': ['py.typed']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    extras_require={
        'dev': DEV_REQUIREMENTS,
    },
    entry_points={
        'console_scripts': [
            'pip-tree=pip_tree.cli:main',
        ],
    },
    python_requires='>=3.10, <4',
)
